# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from anemoi.graphs.describe import GraphDescriptor

from . import Command


class Describe(Command):
    """Describe a graph."""

    internal = True
    timestamp = True

    def add_arguments(self, command_parser):
        command_parser.add_argument("graph_file", help="Path to the graph (a .PT file).")

    def run(self, args):
        kwargs = vars(args)

        GraphDescriptor(kwargs["graph_file"]).describe()


command = Describe



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging

import numpy as np
import torch

from anemoi.graphs.nodes.builders.base import BaseNodeBuilder

LOGGER = logging.getLogger(__name__)


class LatLonNodes(BaseNodeBuilder):
    """Nodes from its latitude and longitude positions (in numpy arrays).

    Attributes
    ----------
    latitudes : list | np.ndarray
        The latitude of the nodes, in degrees.
    longitudes : list | np.ndarray
        The longitude of the nodes, in degrees.

    Methods
    -------
    get_coordinates()
        Get the lat-lon coordinates of the nodes.
    register_nodes(graph, name)
        Register the nodes in the graph.
    register_attributes(graph, name, config)
        Register the attributes in the nodes of the graph specified.
    update_graph(graph, name, attrs_config)
        Update the graph with new nodes and attributes.
    """

    def __init__(
        self,
        latitudes: list[float] | np.ndarray | torch.Tensor,
        longitudes: list[float] | np.ndarray | torch.Tensor,
        name: str,
    ) -> None:
        super().__init__(name)
        self.latitudes = latitudes if isinstance(latitudes, torch.Tensor) else torch.tensor(latitudes)
        self.longitudes = longitudes if isinstance(longitudes, torch.Tensor) else torch.tensor(longitudes)

        assert len(self.latitudes) == len(
            self.longitudes
        ), f"Lenght of latitudes and longitudes must match but {len(self.latitudes)}!={len(self.longitudes)}."
        assert self.latitudes.ndim == 1 or (
            self.latitudes.ndim == 2 and self.latitudes.shape[1] == 1
        ), "latitudes must have shape (N, ) or (N, 1)."
        assert self.longitudes.ndim == 1 or (
            self.longitudes.ndim == 2 and self.longitudes.shape[1] == 1
        ), "longitudes must have shape (N, ) or (N, 1)."

    def get_coordinates(self) -> torch.Tensor:
        """Get the coordinates of the nodes.

        Returns
        -------
        torch.Tensor of shape (num_nodes, 2)
            A 2D tensor with the coordinates, in radians.
        """
        return self.reshape_coords(self.latitudes, self.longitudes)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import pytest
import torch
from torch_geometric.data import HeteroData

from anemoi.graphs.edges.attributes import AttributeFromSourceNode
from anemoi.graphs.edges.attributes import AttributeFromTargetNode
from anemoi.graphs.edges.attributes import DirectionalHarmonics
from anemoi.graphs.edges.attributes import EdgeDirection
from anemoi.graphs.edges.attributes import EdgeLength
from anemoi.graphs.edges.attributes import GaussianDistanceWeights
from anemoi.graphs.edges.attributes import RadialBasisFeatures

TEST_EDGES = ("test_nodes", "to", "test_nodes")


@pytest.mark.parametrize("norm", ["l1", "l2", "unit-max", "unit-std", "unit-range"])
def test_directional_features(graph_nodes_and_edges, norm):
    """Test EdgeDirection compute method."""
    edge_attr_builder = EdgeDirection(norm=norm)
    edge_index = graph_nodes_and_edges[TEST_EDGES].edge_index
    source_nodes = graph_nodes_and_edges[TEST_EDGES[0]]
    target_nodes = graph_nodes_and_edges[TEST_EDGES[2]]
    edge_attr = edge_attr_builder(x=(source_nodes, target_nodes), edge_index=edge_index)
    assert isinstance(edge_attr, torch.Tensor)


@pytest.mark.parametrize("edge_attr_cls", [EdgeLength, GaussianDistanceWeights])
@pytest.mark.parametrize("norm", ["l1", "l2", "unit-max", "unit-std", "unit-range"])
def test_edge_lengths(edge_attr_cls, graph_nodes_and_edges, norm):
    """Test EdgeLength compute method."""
    edge_attr_builder = edge_attr_cls(norm=norm)
    edge_index = graph_nodes_and_edges[TEST_EDGES].edge_index
    source_nodes = graph_nodes_and_edges[TEST_EDGES[0]]
    target_nodes = graph_nodes_and_edges[TEST_EDGES[2]]
    edge_attr = edge_attr_builder(x=(source_nodes, target_nodes), edge_index=edge_index)
    assert isinstance(edge_attr, torch.Tensor)


@pytest.mark.parametrize("attribute_builder_cls", [AttributeFromSourceNode, AttributeFromTargetNode])
def test_edge_attribute_from_node(attribute_builder_cls, graph_nodes_and_edges: HeteroData):
    """Test edge attribute builder fails with unknown nodes."""
    edge_attr_builder = attribute_builder_cls(node_attr_name="mask")
    edge_index = graph_nodes_and_edges[TEST_EDGES].edge_index
    source_nodes = graph_nodes_and_edges[TEST_EDGES[0]]
    target_nodes = graph_nodes_and_edges[TEST_EDGES[2]]
    edge_attr = edge_attr_builder(x=(source_nodes, target_nodes), edge_index=edge_index)
    assert isinstance(edge_attr, torch.Tensor)


@pytest.mark.parametrize("attribute_builder", [EdgeDirection(), EdgeLength()])
def test_fail_edge_features(attribute_builder, graph_nodes_and_edges):
    """Test edge attribute builder fails with unknown nodes."""
    # with pytest.raises(AssertionError):
    edge_index = graph_nodes_and_edges[TEST_EDGES].edge_index
    source_nodes = graph_nodes_and_edges[TEST_EDGES[0]]
    target_nodes = graph_nodes_and_edges[TEST_EDGES[2]]
    attribute_builder(x=(source_nodes, target_nodes), edge_index=edge_index)


def test_radial_basis_features_default(graph_nodes_and_edges):
    """Test RadialBasisFeatures with default parameters (adaptive scaling)."""
    edge_attr_builder = RadialBasisFeatures()
    edge_index = graph_nodes_and_edges[TEST_EDGES].edge_index
    source_nodes = graph_nodes_and_edges[TEST_EDGES[0]]
    target_nodes = graph_nodes_and_edges[TEST_EDGES[2]]
    edge_attr = edge_attr_builder(x=(source_nodes, target_nodes), edge_index=edge_index)

    assert isinstance(edge_attr, torch.Tensor)
    assert edge_attr.shape[0] == edge_index.shape[1]  # num_edges
    assert edge_attr.shape[1] == 5  # default 5 centers

    # Check per-target-node per-RBF-center normalization
    # Within each RBF center, edges to same target should sum to ~1
    target_indices = edge_index[1]  # Target nodes for each edge
    for center_idx in range(edge_attr.shape[1]):
        for target_node in target_indices.unique():
            mask = target_indices == target_node
            center_sum = edge_attr[mask, center_idx].sum()
            assert torch.isclose(center_sum, torch.tensor(1.0, dtype=edge_attr.dtype), atol=1e-6)


def test_radial_basis_features_global_scale(graph_nodes_and_edges):
    """Test RadialBasisFeatures with global r_scale."""
    edge_attr_builder = RadialBasisFeatures(r_scale=1.0)
    edge_index = graph_nodes_and_edges[TEST_EDGES].edge_index
    source_nodes = graph_nodes_and_edges[TEST_EDGES[0]]
    target_nodes = graph_nodes_and_edges[TEST_EDGES[2]]
    edge_attr = edge_attr_builder(x=(source_nodes, target_nodes), edge_index=edge_index)

    assert isinstance(edge_attr, torch.Tensor)
    assert edge_attr.shape[0] == edge_index.shape[1]

    # Check per-target-node per-RBF-center normalization
    target_indices = edge_index[1]
    for center_idx in range(edge_attr.shape[1]):
        for target_node in target_indices.unique():
            mask = target_indices == target_node
            center_sum = edge_attr[mask, center_idx].sum()
            assert torch.isclose(center_sum, torch.tensor(1.0, dtype=edge_attr.dtype), atol=1e-6)


def test_radial_basis_features_custom_centers(graph_nodes_and_edges):
    """Test RadialBasisFeatures with custom centers."""
    custom_centers = [0.0, 0.33, 0.67, 1.0]
    edge_attr_builder = RadialBasisFeatures(centers=custom_centers, sigma=0.15)
    edge_index = graph_nodes_and_edges[TEST_EDGES].edge_index
    source_nodes = graph_nodes_and_edges[TEST_EDGES[0]]
    target_nodes = graph_nodes_and_edges[TEST_EDGES[2]]
    edge_attr = edge_attr_builder(x=(source_nodes, target_nodes), edge_index=edge_index)

    assert isinstance(edge_attr, torch.Tensor)
    assert edge_attr.shape[1] == len(custom_centers)

    # Check per-target-node per-RBF-center normalization
    target_indices = edge_index[1]
    for center_idx in range(edge_attr.shape[1]):
        for target_node in target_indices.unique():
            mask = target_indices == target_node
            center_sum = edge_attr[mask, center_idx].sum()
            assert torch.isclose(center_sum, torch.tensor(1.0, dtype=edge_attr.dtype), atol=1e-6)


def test_radial_basis_features_epsilon(graph_nodes_and_edges):
    """Test RadialBasisFeatures with custom epsilon."""
    edge_attr_builder = RadialBasisFeatures(epsilon=1e-8)
    edge_index = graph_nodes_and_edges[TEST_EDGES].edge_index
    source_nodes = graph_nodes_and_edges[TEST_EDGES[0]]
    target_nodes = graph_nodes_and_edges[TEST_EDGES[2]]
    edge_attr = edge_attr_builder(x=(source_nodes, target_nodes), edge_index=edge_index)

    assert isinstance(edge_attr, torch.Tensor)
    # Should not crash with division by zero


def test_directional_harmonics_default(graph_nodes_and_edges):
    """Test DirectionalHarmonics with default parameters."""
    edge_attr_builder = DirectionalHarmonics()
    edge_index = graph_nodes_and_edges[TEST_EDGES].edge_index
    source_nodes = graph_nodes_and_edges[TEST_EDGES[0]]
    target_nodes = graph_nodes_and_edges[TEST_EDGES[2]]
    edge_attr = edge_attr_builder(x=(source_nodes, target_nodes), edge_index=edge_index)

    assert isinstance(edge_attr, torch.Tensor)
    assert edge_attr.shape[0] == edge_index.shape[1]  # num_edges
    assert edge_attr.shape[1] == 2 * 3  # default order=3 -> 2*order features


@pytest.mark.parametrize("order", [1, 2, 3, 5])
def test_directional_harmonics_order(graph_nodes_and_edges, order):
    """Test DirectionalHarmonics with different orders."""
    edge_attr_builder = DirectionalHarmonics(order=order)
    edge_index = graph_nodes_and_edges[TEST_EDGES].edge_index
    source_nodes = graph_nodes_and_edges[TEST_EDGES[0]]
    target_nodes = graph_nodes_and_edges[TEST_EDGES[2]]
    edge_attr = edge_attr_builder(x=(source_nodes, target_nodes), edge_index=edge_index)

    assert isinstance(edge_attr, torch.Tensor)
    assert edge_attr.shape[1] == 2 * order  # sin and cos for each order


@pytest.mark.parametrize("norm", ["l1", "l2", "unit-max", "unit-std", "unit-range"])
def test_directional_harmonics_with_norm(graph_nodes_and_edges, norm):
    """Test DirectionalHarmonics with different normalization methods."""
    edge_attr_builder = DirectionalHarmonics(order=2, norm=norm)
    edge_index = graph_nodes_and_edges[TEST_EDGES].edge_index
    source_nodes = graph_nodes_and_edges[TEST_EDGES[0]]
    target_nodes = graph_nodes_and_edges[TEST_EDGES[2]]
    edge_attr = edge_attr_builder(x=(source_nodes, target_nodes), edge_index=edge_index)

    assert isinstance(edge_attr, torch.Tensor)
    assert edge_attr.shape[1] == 2 * 2  # order=2


def test_directional_harmonics_values(graph_nodes_and_edges):
    """Test that DirectionalHarmonics produces reasonable values."""
    edge_attr_builder = DirectionalHarmonics(order=1, norm=None)
    edge_index = graph_nodes_and_edges[TEST_EDGES].edge_index
    source_nodes = graph_nodes_and_edges[TEST_EDGES[0]]
    target_nodes = graph_nodes_and_edges[TEST_EDGES[2]]
    edge_attr = edge_attr_builder(x=(source_nodes, target_nodes), edge_index=edge_index)

    # For order=1, features are [sin(ψ), cos(ψ)]
    # sin²(ψ) + cos²(ψ) = 1
    sin_vals = edge_attr[:, 0]
    cos_vals = edge_attr[:, 1]
    norms = sin_vals**2 + cos_vals**2

    assert torch.allclose(norms, torch.ones_like(norms), atol=1e-6)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import numpy as np
import torch

from anemoi.graphs.generate.icon_mesh import ICONCellDataGrid
from anemoi.graphs.generate.icon_mesh import ICONMultiMesh
from anemoi.graphs.nodes.builders.base import BaseNodeBuilder


class BaseICONNodeBuilder(BaseNodeBuilder):
    """Base ICON Node Builder."""

    icon_node_class: type[ICONCellDataGrid] | type[ICONMultiMesh]

    def __init__(self, name: str, grid_filename: str, max_level: int) -> None:
        self.icon_nodes = self.icon_node_class(icon_grid_filename=grid_filename, max_level=max_level)
        super().__init__(name)
        self.hidden_attributes = BaseNodeBuilder.hidden_attributes | {"icon_nodes"}

    def get_coordinates(self) -> torch.Tensor:
        return torch.from_numpy(self.icon_nodes.nodeset.gc_vertices.astype(np.float32)).fliplr()


class ICONMultiMeshNodes(BaseICONNodeBuilder):
    """Processor mesh based on an ICON grid."""

    icon_node_class = ICONMultiMesh


class ICONCellGridNodes(BaseICONNodeBuilder):
    """Data mesh based on an ICON grid."""

    icon_node_class = ICONCellDataGrid



# (C) Copyright 2024-2025 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import logging

from pydantic import BaseModel as PydanticBaseModel
from pydantic import Field
from pydantic import model_validator
from typing_extensions import Self

from anemoi.utils.schemas import BaseModel

from .edge_attributes_schemas import EdgeAttributeSchema  # noqa: TC001
from .edge_schemas import EdgeBuilderSchemas  # noqa: TC001
from .node_attributes_schemas import NodeAttributeSchemas  # noqa: TC001
from .node_schemas import NodeBuilderSchemas  # noqa: TC001
from .post_processors import ProcessorSchemas  # noqa: TC001

LOGGER = logging.getLogger(__name__)


class NodeSchema(BaseModel):
    node_builder: NodeBuilderSchemas
    "Node builder schema."
    attributes: dict[str, NodeAttributeSchemas] | None = None
    "Dictionary of attributes with names as keys and anemoi.graphs.nodes.attributes objects as values."


class EdgeSchema(BaseModel):
    source_name: str = Field(examples=["data", "hidden"])
    "Source of the edges."
    target_name: str = Field(examples=["data", "hidden"])
    "Target of the edges."
    edge_builders: list[EdgeBuilderSchemas]
    "Edge builder schema."
    attributes: dict[str, EdgeAttributeSchema]
    "Dictionary of attributes with names as keys and anemoi.graphs.edges.attributes objects as values."


class BaseGraphSchema(PydanticBaseModel):
    nodes: dict[str, NodeSchema] | None = Field(default=None)
    "Nodes schema for all types of nodes (ex. data, hidden)."
    edges: list[EdgeSchema] | None = Field(default=None)
    "List of edges schema."
    overwrite: bool = Field(example=True)
    "whether to overwrite existing graph file. Default to True."
    post_processors: list[ProcessorSchemas] = Field(default_factory=list)
    # TODO(Helen): Needs to be adjusted for more complex graph setups

    @model_validator(mode="after")
    def check_if_nodes_edges_present_if_overwrite(self) -> Self:
        if self.overwrite and ("nodes" not in self.model_fields_set or "edges" not in self.model_fields_set):
            msg = "If overwrite is True, nodes and edges must be provided."
            raise ValueError(msg)
        return self



# (C) Copyright 2024-2025 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


from typing import Annotated
from typing import Literal

from pydantic import Field
from pydantic import PositiveFloat
from pydantic import PositiveInt
from pydantic import model_validator

from anemoi.utils.schemas import BaseModel


class KNNEdgeSchema(BaseModel):
    target_: Literal["anemoi.graphs.edges.KNNEdges", "anemoi.graphs.edges.ReversedKNNEdges"] = Field(
        ..., alias="_target_"
    )
    "KNN based edges implementation from anemoi.graphs.edges."
    num_nearest_neighbours: PositiveInt = Field(example=3)
    "Number of nearest neighbours. Default to 3."
    source_mask_attr_name: str | None = Field(default=None, examples=["boundary_mask"])
    "Mask to apply to source nodes of the edges. Default to None."
    target_mask_attr_name: str | None = Field(default=None, examples=["boundary_mask"])
    "Mask to apply to target nodes of the edges. Default to None."


class CutoffEdgeSchema(BaseModel):
    target_: Literal["anemoi.graphs.edges.CutOffEdges", "anemoi.graphs.edges.ReversedCutOffEdges"] = Field(
        ..., alias="_target_"
    )
    "Cut-off based edges implementation from anemoi.graphs.edges."
    cutoff_factor: PositiveFloat | None = Field(default=None, example=0.6)
    "Factor to multiply the grid reference distance to get the cut-off radius. Mutually exclusive with cutoff_distance_km."
    cutoff_distance_km: PositiveFloat | None = Field(default=None, example=500.0)
    "Cutoff radius in kilometers. Mutually exclusive with cutoff_factor."
    source_mask_attr_name: str | None = Field(default=None, examples=["boundary_mask"])
    "Mask to apply to source nodes of the edges. Default to None."
    target_mask_attr_name: str | None = Field(default=None, examples=["boundary_mask"])
    "Mask to apply to target nodes of the edges. Default to None."
    max_num_neighbours: PositiveInt = Field(default=64, example=64)
    "Maximum number of nearest neighbours to consider when building edges. Default to 64."

    @model_validator(mode="after")
    def validate_cutoff_params(self):
        """Validate that exactly one of cutoff_factor or cutoff_distance_km is provided."""
        cutoff_factor = self.cutoff_factor
        cutoff_distance_km = self.cutoff_distance_km

        if cutoff_factor is None and cutoff_distance_km is None:
            raise ValueError("Either cutoff_factor or cutoff_distance_km must be provided.")
        if cutoff_factor is not None and cutoff_distance_km is not None:
            raise ValueError("cutoff_factor and cutoff_distance_km are mutually exclusive. Provide only one.")

        return self


class MultiScaleEdgeSchema(BaseModel):
    target_: Literal["anemoi.graphs.edges.MultiScaleEdges"] = Field(
        "anemoi.graphs.edges.MultiScaleEdges",
        alias="_target_",
    )
    "Multi-scale edges implementation from anemoi.graphs.edges."
    x_hops: PositiveInt = Field(example=1)
    "Number of hops (in the refined icosahedron) between two nodes to connect them with an edge. Default to 1."
    scale_resolutions: PositiveInt | list[PositiveInt] | None = Field(examples=[1, 2, 3, 4, 5])
    "Specifies the resolution scales for computing the hop neighbourhood."
    source_mask_attr_name: str | None = Field(default=None, examples=["boundary_mask"])
    "Mask to apply to source nodes of the edges. Default to None."
    target_mask_attr_name: str | None = Field(default=None, examples=["boundary_mask"])
    "Mask to apply to target nodes of the edges. Default to None."


class HEALPixMultiScaleEdgesSchema(BaseModel):
    target_: Literal["anemoi.graphs.edges.HEALPixMultiScaleEdges"] = Field(..., alias="_target_")
    "HEALPix multi-scale edges implementation from anemoi.graphs.edges."
    scale_resolutions: PositiveInt | list[PositiveInt] | None = Field(examples=[1, 2, 3, 4, 5])
    "Specifies the resolution scales for computing the hop neighbourhood."
    source_mask_attr_name: str | None = Field(default=None, examples=["boundary_mask"])
    "Mask to apply to source nodes of the edges. Default to None."
    target_mask_attr_name: str | None = Field(default=None, examples=["boundary_mask"])
    "Mask to apply to target nodes of the edges. Default to None."


class ICONTopologicalEdgeSchema(BaseModel):
    target_: Literal[
        "anemoi.graphs.edges.ICONTopologicalProcessorEdges",
        "anemoi.graphs.edges.ICONTopologicalEncoderEdges",
        "anemoi.graphs.edges.ICONTopologicalDecoderEdges",
    ] = Field("anemoi.graphs.edges.ICONTopologicalProcessorEdges", alias="_target_")
    source_mask_attr_name: str | None = Field(default=None, examples=["boundary_mask"])
    "Mask to apply to source nodes of the edges. Default to None."
    target_mask_attr_name: str | None = Field(default=None, examples=["boundary_mask"])
    "Mask to apply to target nodes of the edges. Default to None."


class EdgeAttributeSchema(BaseModel):
    target_: Literal["anemoi.graphs.edges.attributes.EdgeLength", "anemoi.graphs.edges.attributes.EdgeDirection"] = (
        Field("anemoi.graphs.edges.attributes.EdgeLength", alias="_target_")
    )
    "Edge attributes object from anemoi.graphs.edges."
    norm: Literal["unit-max", "l1", "l2", "unit-sum", "unit-std"] = Field(example="unit-std")
    "Normalisation method applied to the edge attribute."


EdgeBuilderSchemas = Annotated[
    KNNEdgeSchema | CutoffEdgeSchema | MultiScaleEdgeSchema | HEALPixMultiScaleEdgesSchema | ICONTopologicalEdgeSchema,
    Field(discriminator="target_"),
]



# (C) Copyright 2024- Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from abc import ABC
from abc import abstractmethod

import numpy as np
import torch
from torch_geometric.data.storage import NodeStorage

from anemoi.datasets import open_dataset
from anemoi.graphs.nodes.attributes.base_attributes import BooleanBaseNodeAttribute

LOGGER = logging.getLogger(__name__)


class BaseAnemoiDatasetVariable(BooleanBaseNodeAttribute):
    """Base class for computing mask based on a variable in an Anemoi dataset."""

    def __init__(self, variable: str) -> None:
        super().__init__()
        self.variable = variable

    @abstractmethod
    def _get_mask(self, ds) -> np.ndarray: ...

    def _read_data(self, nodes: NodeStorage, **kwargs) -> np.ndarray:
        return open_dataset(nodes["_dataset"], select=self.variable)[0].squeeze()

    def get_raw_values(self, nodes: NodeStorage, **kwargs) -> torch.Tensor:

        assert nodes["node_type"] in [
            "ZarrDatasetNodes",
            "AnemoiDatasetNodes",
        ], f"{self.__class__.__name__} can only be used with AnemoiDatasetNodes."
        ds = self._read_data(nodes)
        return torch.from_numpy(self._get_mask(ds))


class NonmissingAnemoiDatasetVariable(BaseAnemoiDatasetVariable):
    """Mask of valid (not missing) values of an Anemoi dataset variable.

    It reads a variable from an Anemoi dataset and returns a boolean mask of nonmissing values in the first timestep.

    Attributes
    ----------
    variable : str
        Variable to read from the Anemoi dataset.

    Methods
    -------
    compute(self, graph, nodes_name)
        Compute the attribute for each node.
    """

    def __init__(self, variable: str) -> None:
        super().__init__(variable)
        self.variable = variable

    def _get_mask(self, ds) -> np.ndarray:
        return ~np.isnan(ds)


class NonzeroAnemoiDatasetVariable(BaseAnemoiDatasetVariable):
    """Mask of non-zero values of an Anemoi dataset variable.

    Reads a variable from an Anemoi dataset and returns a boolean mask of non-zero values in the first timestep.

    Attributes
    ----------
    variable : str
        Variable to read from the Anemoi dataset.

    Methods
    -------
    compute(self, graph, nodes_name)
        Computer the attribute for each node.
    """

    def __init__(self, variable: str) -> None:
        super().__init__(variable)
        self.variable = variable

    def _get_mask(self, ds) -> np.ndarray:
        return ds != 0


class BaseCombineAnemoiDatasetsMask(BooleanBaseNodeAttribute, ABC):
    """Base class for computing mask based on anemoi-datasets combining operations."""

    grids: list[int] | None = None

    def __init__(self) -> None:
        super().__init__()
        if self.grids is None:
            raise AttributeError(f"{self.__class__.__name__} class must set 'grids' attribute.")

    def get_grid_sizes(self, nodes):
        from anemoi.datasets import open_dataset

        assert "_dataset" in nodes and isinstance(
            nodes["_dataset"], (dict, str)
        ), "The '_dataset' attribute must be a dictionary or string."

        return open_dataset(nodes["_dataset"]).grids

    @staticmethod
    def get_mask_from_grid_sizes(grid_sizes: tuple[int], masked_grids_posisitons: list[int]):
        assert isinstance(masked_grids_posisitons, list), "masked_grids_positions must be a list"
        assert min(masked_grids_posisitons) >= 0, "masked_grids_positions must be non-negative"
        assert max(masked_grids_posisitons) < len(grid_sizes), f"masked_grids_positions must be < {len(grid_sizes)}"
        mask = torch.zeros(sum(grid_sizes), dtype=torch.bool)
        for grid_id in masked_grids_posisitons:
            mask[sum(grid_sizes[:grid_id]) : sum(grid_sizes[: grid_id + 1])] = True
        return mask

    def get_raw_values(self, nodes: NodeStorage, **kwargs) -> torch.Tensor:
        grid_sizes = self.get_grid_sizes(nodes)
        return BaseCombineAnemoiDatasetsMask.get_mask_from_grid_sizes(grid_sizes, self.grids)


class CutOutMask(BaseCombineAnemoiDatasetsMask):
    """Cut out mask.

    It computes a mask for the first dataset in the cutout operation.

    Methods
    -------
    compute(self, graph, nodes_name)
        Compute the attribute for each node.
    """

    def __init__(self) -> None:
        self.grids = [0]  # It sets as true the nodes from the first (index=0) grid
        super().__init__()


class GridsMask(BaseCombineAnemoiDatasetsMask):
    """Grids mask.

    It reads a variable from a Anemoi dataset and returns a boolean mask of nonmissing values in the first timestep.

    Attributes
    ----------
    grids : int | list[int], optional
        Grid positions to set as True. Defaults to 0, which sets True only the nodes from the first dataset.

    Methods
    -------
    compute(self, graph, nodes_name)
        Compute the attribute for each node.
    """

    def __init__(self, grids: int | list[int] = 0) -> None:
        self.grids = [grids] if isinstance(grids, int) else grids
        super().__init__()


class LimitedAreaMask(BooleanBaseNodeAttribute):
    """Limited area mask.

    It adds a mask based on an area of interest. This mask is only defined
    for nodes built with a subclass of `StretchedIcosahedronNodes`.

    Methods
    -------
    compute(self, graph, nodes_name)
        Compute the attribute for each node.
    """

    def get_raw_values(self, nodes: NodeStorage, **kwargs) -> torch.Tensor:
        assert nodes["node_type"] in [
            "StretchedTriNodes"
        ], f"{self.__class__.__name__} can only be used with StretchedIcosahedronNodes."
        lam_mask = nodes["_area_mask_builder"].get_mask(nodes.x)
        return torch.from_numpy(lam_mask)



# (C) Copyright 2025- Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
from typing import Optional

import torch
from torch_geometric.data.storage import NodeStorage

from anemoi.graphs.edges.builders.base import BaseEdgeBuilder

LOGGER = logging.getLogger(__name__)


class HEALPixMultiScaleEdges(BaseEdgeBuilder):
    """HEALPix Multi Scale Edges."""

    def __init__(
        self,
        source_name: str,
        target_name: str,
        scale_resolutions: Optional[int | list[int]] = None,
        **kwargs,
    ):
        super().__init__(source_name, target_name)
        assert source_name == target_name, f"{self.__class__.__name__} requires source and target nodes to be the same."
        if isinstance(scale_resolutions, int):
            assert scale_resolutions > 0, "The scale_resolutions argument only supports positive integers."
            scale_resolutions = list(range(1, scale_resolutions + 1))
        assert not isinstance(scale_resolutions, str), "The scale_resolutions argument is not valid."
        assert (
            scale_resolutions is None or min(scale_resolutions) > 0
        ), "The scale_resolutions argument only supports positive integers."
        self.scale_resolutions = scale_resolutions

    def compute_edge_index(self, source_nodes: NodeStorage, target_nodes: NodeStorage) -> torch.Tensor:
        """Compute the edge index for HEALPix multi scale edges."""
        assert source_nodes.node_type == "HEALPixNodes", f"{self.__class__.__name__} only supports HEALPixNodes."

        from anemoi.graphs.generate.healpix import get_healpix_edgeindex

        scale_resolutions = self.scale_resolutions or list(range(1, source_nodes["_resolution"] + 1))
        edges_index, prev_res = None, None
        for res in list(sorted(scale_resolutions)):
            new_edge_index = get_healpix_edgeindex(res)
            LOGGER.debug(f"Resolution: {res}, Edge index shape: {new_edge_index.shape}")
            if edges_index is None:
                edges_index = new_edge_index
            else:
                edges_index = torch.cat([4 ** (res - prev_res) * edges_index, new_edge_index], dim=1)
            prev_res = res

        return edges_index



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from abc import ABC
from abc import abstractmethod

import torch
from torch_geometric.data.storage import NodeStorage
from torch_geometric.nn.conv import MessagePassing
from torch_geometric.typing import Adj
from torch_geometric.typing import PairTensor
from torch_geometric.typing import Size
from torch_geometric.utils import scatter

from anemoi.graphs.edges.directional import compute_directions
from anemoi.graphs.normalise import NormaliserMixin
from anemoi.graphs.utils import NodesAxis
from anemoi.graphs.utils import get_distributed_device
from anemoi.graphs.utils import haversine_distance

LOGGER = logging.getLogger(__name__)


class BaseEdgeAttributeBuilder(MessagePassing, NormaliserMixin, ABC):
    """Base class for edge attribute builders."""

    node_attr_name: str = None
    norm_by_group: bool = False

    def __init__(self, norm: str | None = None, dtype: str = "float32") -> None:
        super().__init__()
        self.norm = norm
        self.dtype = dtype
        self.device = get_distributed_device()
        if self.node_attr_name is None:
            error_msg = f"Class {self.__class__.__name__} must define 'node_attr_name' either as a class attribute or in __init__"
            raise TypeError(error_msg)

    def subset_node_information(self, source_nodes: NodeStorage, target_nodes: NodeStorage) -> PairTensor:
        if self.node_attr_name in source_nodes:
            source_nodes_data = source_nodes[self.node_attr_name].to(self.device)
        else:
            source_nodes_data = None
            LOGGER.warning("The attribute %s is not in the source nodes.", self.node_attr_name)

        if self.node_attr_name in target_nodes:
            target_nodes_data = target_nodes[self.node_attr_name].to(self.device)
        else:
            target_nodes_data = None
            LOGGER.warning("The attribute %s is not in the target nodes.", self.node_attr_name)

        return source_nodes_data, target_nodes_data

    def forward(self, x: tuple[NodeStorage, NodeStorage], edge_index: Adj, size: Size = None) -> torch.Tensor:
        x = self.subset_node_information(*x)
        return self.propagate(edge_index.to(self.device), x=x, size=size)

    @abstractmethod
    def compute(self, x_i: torch.Tensor, x_j: torch.Tensor) -> torch.Tensor: ...

    def message(self, x_i: torch.Tensor, x_j: torch.Tensor) -> torch.Tensor:
        edge_features = self.compute(x_i, x_j)

        if edge_features.ndim == 1:
            edge_features = edge_features.unsqueeze(-1)

        return edge_features

    def aggregate(self, edge_features: torch.Tensor, index: torch.Tensor, ptr=None, dim_size=None) -> torch.Tensor:
        return self.normalise(edge_features, index, dim_size)


class BasePositionalBuilder(BaseEdgeAttributeBuilder, ABC):
    node_attr_name: str = "x"
    _idx_lat: int = 0
    _idx_lon: int = 1


class EdgeLength(BasePositionalBuilder):
    """Computes edge length for bipartite graphs."""

    def compute(self, x_i: torch.Tensor, x_j: torch.Tensor) -> torch.Tensor:
        edge_length = haversine_distance(x_i, x_j)
        return edge_length


class EdgeDirection(BasePositionalBuilder):
    """Computes edge direction for bipartite graphs."""

    def compute(self, x_i: torch.Tensor, x_j: torch.Tensor) -> torch.Tensor:
        edge_dirs = compute_directions(source_coords=x_j, target_coords=x_i)
        return edge_dirs


class DirectionalHarmonics(EdgeDirection):
    """Computes directional harmonics from edge directions.

    Builds directional harmonics [sin(mψ), cos(mψ)]_{m=1..order} from per-edge
    2D directions (dx, dy). Returns shape [N, 2*order].

    Attributes
    ----------
    order : int
        The maximum order of harmonics to compute.
    norm : str | None
        Normalisation method. Options: None, "l1", "l2", "unit-max", "unit-range", "unit-std".

    Methods
    -------
    compute(x_i, x_j)
        Compute directional harmonics from edge directions.
    """

    def __init__(self, order: int = 3, norm: str | None = None, dtype: str = "float32") -> None:
        self.order = order
        super().__init__(norm=norm, dtype=dtype)

    def compute(self, x_i: torch.Tensor, x_j: torch.Tensor) -> torch.Tensor:
        # Get the 2D direction vectors [dx, dy]
        edge_dirs = compute_directions(source_coords=x_j, target_coords=x_i)

        # Compute the angle ψ from the direction vectors
        psi = torch.atan2(edge_dirs[:, 1], edge_dirs[:, 0])  # atan2(dy, dx)

        # Build harmonics: [sin(ψ), cos(ψ), sin(2ψ), cos(2ψ), ..., sin(order*ψ), cos(order*ψ)]
        harmonics = []
        for m in range(1, self.order + 1):
            harmonics.append(torch.sin(m * psi))
            harmonics.append(torch.cos(m * psi))

        # Stack into shape [N, 2*order]
        return torch.stack(harmonics, dim=1)


class Azimuth(BasePositionalBuilder):
    """Compute the azimuth of the edge.

    Attributes
    ----------
    norm : str | None
        Normalisation method. Options: None, "l1", "l2", "unit-max", "unit-range", "unit-std".
    invert : bool
        Whether to invert the edge lengths, i.e. 1 - edge_length. Defaults to False.

    Methods
    -------
    compute(x_i, x_j)
        Compute edge lengths attributes.

    References
    ----------
    - https://www.movable-type.co.uk/scripts/latlong.html
    """

    def compute(self, x_i: torch.Tensor, x_j: torch.Tensor) -> torch.Tensor:
        # Forward bearing. x_i, x_j must be radians.
        a11 = torch.cos(x_i[:, self._idx_lat]) * torch.sin(x_j[:, self._idx_lat])
        a12 = (
            torch.sin(x_i[:, self._idx_lat])
            * torch.cos(x_j[:, self._idx_lat])
            * torch.cos(x_j[..., self._idx_lon] - x_i[..., self._idx_lon])
        )
        a1 = a11 - a12
        a2 = torch.sin(x_j[..., self._idx_lon] - x_i[..., self._idx_lon]) * torch.cos(x_j[:, self._idx_lat])
        edge_dirs = torch.atan2(a2, a1)

        return edge_dirs


class BaseBooleanEdgeAttributeBuilder(BaseEdgeAttributeBuilder, ABC):
    """Base class for boolean edge attributes."""

    def __init__(self) -> None:
        super().__init__(norm=None, dtype="bool")


class BaseEdgeAttributeFromNodeBuilder(BaseBooleanEdgeAttributeBuilder, ABC):
    """Base class for propagating an attribute from the nodes to the edges."""

    nodes_axis: NodesAxis | None = None

    def __init__(self, node_attr_name: str) -> None:
        self.node_attr_name = node_attr_name
        super().__init__()
        if self.nodes_axis is None:
            raise AttributeError(f"{self.__class__.__name__} class must set 'nodes_axis' attribute.")

    def compute(self, x_i: torch.Tensor, x_j: torch.Tensor) -> torch.Tensor:
        node_attr = (x_j, x_i)[self.nodes_axis.value]
        assert (
            node_attr is not None
        ), f"The node attribute specified for {self.node_attr_name} cannot be found in the nodes."
        return node_attr


class AttributeFromSourceNode(BaseEdgeAttributeFromNodeBuilder):
    """Copy an attribute of the source node to the edge."""

    nodes_axis = NodesAxis.SOURCE


class AttributeFromTargetNode(BaseEdgeAttributeFromNodeBuilder):
    """Copy an attribute of the target node to the edge."""

    nodes_axis = NodesAxis.TARGET


class RadialBasisFeatures(EdgeLength):
    """Radial basis features from edge distances using Gaussian RBFs.

    Computes Gaussian radial basis function features from normalized great-circle distances:
    phi_r = [exp(-((α - c)/σ)²) for c in centers], where α = r_ij / r_scale.

    Provides RBF features via per-node adaptive scaling.
    By default, each destination node's edges are normalized by that node's maximum edge length.
    RBF features are normalized per target node per RBF center: within each RBF center,
    all edges pointing to the same target node have values that sum to 1 (L1 norm).

    Parameters
    ----------
    r_scale : float | None, optional
        Global scale factor for normalizing distances. Default is None.
        If None: Use per-node adaptive scaling (max edge length per destination node).
        If float: Use global scale for all nodes.
    centers : list of float, optional
        RBF center positions along normalized distance axis [0, 1].
        Default is [0.0, 0.25, 0.5, 0.75, 1.0].
    sigma : float, optional
        Width (standard deviation) of Gaussian RBF functions. Default is 0.2.
        Controls how localized each basis function is around its center.
    epsilon : float, optional
        Small constant to avoid division by zero. Default is 1e-10.
    dtype : str, optional
        Data type for computations. Default is "float32".

    Note
    ----
    RBF features are normalized per target node per RBF center.
    Within each RBF center, all edges to the same target node sum to 1.

    Methods
    -------
    compute(x_i, x_j)
        Compute raw edge distances (RBF computation happens in aggregate).
    aggregate(edge_features, index, ptr, dim_size)
        Compute RBF features with adaptive scaling and per-target-node normalization.

    Examples
    --------
    # Default: per-node adaptive scaling with grouped normalization
    rbf = RadialBasisFeatures()

    # To use global scale
    rbf_global = RadialBasisFeatures(r_scale=1.0)

    # Custom RBF centers and width
    rbf_custom = RadialBasisFeatures(centers=[0.0, 0.33, 0.67, 1.0], sigma=0.15)

    Notes
    -----
    - Closer edges → higher values at low-distance centers (0.0, 0.25)
    - Farther edges → higher values at high-distance centers (0.75, 1.0)
    """

    norm_by_group: bool = True  # normalise the RBF features per destination node

    def __init__(
        self,
        r_scale: float | None = None,
        centers: list[float] | None = None,
        sigma: float = 0.2,
        norm: str = "l1",
        epsilon: float = 1e-10,
        dtype: str = "float32",
    ) -> None:
        self.epsilon = epsilon
        self.r_scale = r_scale

        if self.r_scale is not None and self.r_scale < self.epsilon:
            LOGGER.warning(
                "r_scale (%f) is too small (< epsilon=%f). Clamping to epsilon to avoid division by zero.",
                self.r_scale,
                self.epsilon,
            )
            self.r_scale = self.epsilon

        self.centers = centers if centers is not None else [0.0, 0.25, 0.5, 0.75, 1.0]

        # Normalize centers if using global scaling
        if self.r_scale is not None:
            self.centers = [c / self.r_scale for c in self.centers]

        # Check that centers are in the range [0, 1]
        assert all(
            0.0 <= c <= 1.0 for c in self.centers
        ), f"RBF centers must be in range [0, 1] (or [0, r_scale] if r_scale is set). Got centers: {centers}, r_scale: {r_scale}"

        self.sigma = sigma
        super().__init__(norm=norm, dtype=dtype)

    def aggregate(self, edge_features: torch.Tensor, index: torch.Tensor, ptr=None, dim_size=None) -> torch.Tensor:
        """Aggregate edge features with per-node scaling and per-target-node normalization.

        Parameters
        ----------
        edge_features : torch.Tensor
            Raw edge distances, shape [num_edges] or [num_edges, 1]
        index : torch.Tensor
            Destination node index for each edge
        ptr : optional
            CSR pointer (not used)
        dim_size : int, optional
            Number of destination nodes

        Returns
        -------
        torch.Tensor
            RBF features, shape [num_edges, num_centers].
            Normalized per target node per RBF center .
        """
        # Ensure edge_features is 1D
        if edge_features.ndim == 2:
            edge_features = edge_features.squeeze(-1)

        # Compute scale factor per destination node
        if self.r_scale is None:
            # Per-node max edge length scaling
            max_dists = scatter(edge_features, index.long(), dim=0, dim_size=dim_size, reduce="max")

            # Clamp to epsilon to avoid division by zero
            max_dists = torch.clamp(max_dists, min=self.epsilon)

            # Broadcast to each edge
            scales = max_dists[index]
            alpha = edge_features / scales  # Normalized distance [0, 1]
        else:
            # Global scaling
            scales = torch.full_like(edge_features, self.r_scale)
            alpha = edge_features / scales  # Scaled distance [0, max_edge/r_scale]

        # Compute Gaussian RBF for each center
        rbf_features = []
        for center in self.centers:
            rbf = torch.exp(-(((alpha - center) / self.sigma) ** 2))
            rbf_features.append(rbf)

        rbf_features = torch.stack(rbf_features, dim=1)

        # Within each RBF center, normalise edges to the same target node
        rbf_features = self.normalise(rbf_features, index, dim_size)

        return rbf_features


class GaussianDistanceWeights(EdgeLength):
    """Gaussian distance weights."""

    norm_by_group: bool = True  # normalise the gaussian weights by target node

    def __init__(self, sigma: float = 1.0, norm: str = "l1", **kwargs) -> None:
        self.sigma = sigma
        super().__init__(norm=norm)

    def compute(self, x_i: torch.Tensor, x_j: torch.Tensor) -> torch.Tensor:
        dists = super().compute(x_i, x_j)
        gaussian_weights = torch.exp(-(dists**2) / (2 * self.sigma**2))
        return gaussian_weights



# (C) Copyright 2024-2025 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import logging
from typing import Literal

from pydantic import Field

from anemoi.graphs.schemas.normalise import ImplementedNormalisationSchema
from anemoi.utils.schemas import BaseModel

LOGGER = logging.getLogger(__name__)


class PlanarAreaWeightSchema(BaseModel):
    target_: Literal[
        "anemoi.graphs.nodes.attributes.PlanarAreaWeights",
        "anemoi.graphs.nodes.attributes.UniformWeights",
        "anemoi.graphs.nodes.attributes.CosineLatWeightedAttribute",
        "anemoi.graphs.nodes.attributes.IsolatitudeAreaWeights",
    ] = Field(..., alias="_target_")
    "Implementation of the area of the nodes as the weights from anemoi.graphs.nodes.attributes."
    norm: ImplementedNormalisationSchema = Field(example="unit-max")
    "Normalisation of the weights."


class MaskedPlanarAreaWeightsSchema(BaseModel):
    target_: Literal["anemoi.graphs.nodes.attributes.MaskedPlanarAreaWeights"] = Field(..., alias="_target_")
    "Implementation of the area of the nodes as the weights from anemoi.graphs.nodes.attributes."
    mask_node_attr_name: str = Field(examples="cutout_mask")
    "Attribute name to mask the area weights."
    norm: ImplementedNormalisationSchema = Field(example="unit-max")
    "Normalisation of the weights."


class SphericalAreaWeightSchema(BaseModel):
    target_: Literal["anemoi.graphs.nodes.attributes.SphericalAreaWeights"] = Field(..., alias="_target_")
    "Implementation of the 3D area of the nodes as the weights from anemoi.graphs.nades.attributes."
    norm: ImplementedNormalisationSchema = Field(example="unit-max")
    "Normalisation of the weights."
    fill_value: float = Field(example=0)
    "Value to fill the empty regions."


class CutOutMaskSchema(BaseModel):
    target_: Literal["anemoi.graphs.nodes.attributes.CutOutMask", "anemoi.graphs.nodes.attributes.LimitedAreaMask"] = (
        Field(..., alias="_target_")
    )
    "Implementation of the area masks from anemoi.graphs.nodes.attributes."


class GridsMaskSchema(BaseModel):
    target_: Literal["anemoi.graphs.nodes.attributes.GridsMask"] = Field(..., alias="_target_")
    "Implementation of the grids mask from anemoi.graphs.nodes.attributes."
    grids: list[int] | int = Field(examples=[0, [0]])
    "Position of the grids to consider as True."


class NonmissingAnemoiDatasetVariableSchema(BaseModel):
    target_: Literal["anemoi.graphs.nodes.attributes.NonmissingAnemoiDatasetVariable"] = Field(..., alias="_target_")
    (
        "Implementation of a mask from the nonmissing values of a anemoi-datasets variable "
        "from anemoi.graphs.nodes.attributes."
    )
    variable: str
    "The anemoi-datasets variable to use."


SingleAttributeSchema = (
    PlanarAreaWeightSchema
    | MaskedPlanarAreaWeightsSchema
    | SphericalAreaWeightSchema
    | CutOutMaskSchema
    | GridsMaskSchema
    | NonmissingAnemoiDatasetVariableSchema
)


class BooleanOperationSchema(BaseModel):
    target_: Literal[
        "anemoi.graphs.nodes.attributes.BooleanNot",
        "anemoi.graphs.nodes.attributes.BooleanAndMask",
        "anemoi.graphs.nodes.attributes.BooleanOrMask",
    ] = Field(..., alias="_target_")
    "Implementation of boolean masks from anemoi.graphs.nodes.attributes"
    masks: str | SingleAttributeSchema | list[str | SingleAttributeSchema]


NodeAttributeSchemas = SingleAttributeSchema | BooleanOperationSchema



# (C) Copyright 2024-2025 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


from enum import Enum
from typing import Literal

from pydantic import Field

from anemoi.graphs.schemas.normalise import ImplementedNormalisationSchema
from anemoi.utils.schemas import BaseModel


class ImplementedEdgeAttributeSchema(str, Enum):
    edge_length = "anemoi.graphs.edges.attributes.EdgeLength"
    edge_dirs = "anemoi.graphs.edges.attributes.EdgeDirection"
    directional_harmonics = "anemoi.graphs.edges.attributes.DirectionalHarmonics"
    azimuth = "anemoi.graphs.edges.attributes.Azimuth"
    gaussian_weights = "anemoi.graphs.edges.attributes.GaussianDistanceWeights"
    radial_basis_features = "anemoi.graphs.edges.attributes.RadialBasisFeatures"


class BaseEdgeAttributeSchema(BaseModel):
    target_: ImplementedEdgeAttributeSchema = Field(..., alias="_target_")
    "Edge attribute builder object from anemoi.graphs.edges.attributes"
    norm: ImplementedNormalisationSchema = Field(example="unit-std")
    "Normalisation method applied to the edge attribute."


class EdgeAttributeFromNodeSchema(BaseModel):
    target_: Literal[
        "anemoi.graphs.edges.attributes.AttributeFromSourceNode",
        "anemoi.graphs.edges.attributes.AttributeFromTargetNode",
    ] = Field(..., alias="_target_")
    "Edge attributes from node attribute"
    norm: ImplementedNormalisationSchema = Field(example="unit-std")
    "Normalisation method applied to the edge attribute."


class DirectionalHarmonicsSchema(BaseModel):
    target_: Literal["anemoi.graphs.edges.attributes.DirectionalHarmonics"] = Field(..., alias="_target_")
    "Directional harmonics from edge directions"
    order: int = Field(default=3, description="Maximum order of harmonics to compute")
    norm: ImplementedNormalisationSchema | None = Field(default=None, description="Normalization method")
    dtype: str = Field(default="float32", description="Data type for computations")


class RadialBasisFeaturesSchema(BaseModel):
    target_: Literal["anemoi.graphs.edges.attributes.RadialBasisFeatures"] = Field(..., alias="_target_")
    "Radial basis function features from edge distances"
    r_scale: float | None = Field(default=None, description="Global scale factor (None for adaptive per-node scaling)")
    centers: list[float] | None = Field(default=None, description="RBF center positions [0, 1]")
    sigma: float = Field(default=0.2, description="Width of Gaussian RBF functions")
    epsilon: float = Field(default=1e-10, description="Small constant to avoid division by zero")
    dtype: str = Field(default="float32", description="Data type for computations")


EdgeAttributeSchema = (
    BaseEdgeAttributeSchema | EdgeAttributeFromNodeSchema | DirectionalHarmonicsSchema | RadialBasisFeaturesSchema
)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from abc import ABC
from abc import abstractmethod

import networkx as nx
import numpy as np
import torch
from torch_geometric.data import HeteroData

from anemoi.graphs.generate.masks import KNNAreaMaskBuilder
from anemoi.graphs.nodes.builders.base import BaseNodeBuilder

LOGGER = logging.getLogger(__name__)


class IcosahedralNodes(BaseNodeBuilder, ABC):
    """Nodes based on iterative refinements of an icosahedron.

    Attributes
    ----------
    resolution : list[int] | int
        Refinement level of the mesh.
    """

    def __init__(
        self,
        resolution: int | list[int],
        name: str,
    ) -> None:
        if isinstance(resolution, int):
            self.resolutions = list(range(resolution + 1))
        else:
            self.resolutions = resolution

        super().__init__(name)
        self.hidden_attributes = BaseNodeBuilder.hidden_attributes | {
            "resolutions",
            "nx_graph",
            "node_ordering",
        }
        if not hasattr(self, "multi_scale_edge_cls"):
            raise AttributeError("Classes inheriting from IcosahedralNodes must set 'multi_scale_edge_cls' attribute.")

    def get_coordinates(self) -> torch.Tensor:
        """Get the coordinates of the nodes.

        Returns
        -------
        torch.Tensor of shape (num_nodes, 2)
            A 2D tensor with the coordinates, in radians.
        """
        self.nx_graph, coords_rad, self.node_ordering = self.create_nodes()
        return torch.tensor(coords_rad[self.node_ordering], dtype=torch.float32)

    @abstractmethod
    def create_nodes(self) -> tuple[nx.DiGraph, np.ndarray, list[int]]: ...


class LimitedAreaIcosahedralNodes(IcosahedralNodes, ABC):
    """Nodes based on iterative refinements of an icosahedron using an area of interest.

    Attributes
    ----------
    area_mask_builder : KNNAreaMaskBuilder
        The area of interest mask builder.
    """

    def __init__(
        self,
        resolution: int | list[int],
        reference_node_name: str,
        name: str,
        mask_attr_name: str | None = None,
        margin_radius_km: float = 100.0,
    ) -> None:

        super().__init__(resolution, name)
        self.hidden_attributes = self.hidden_attributes | {"area_mask_builder"}

        self.area_mask_builder = KNNAreaMaskBuilder(reference_node_name, margin_radius_km, mask_attr_name)

    def register_nodes(self, graph: HeteroData) -> None:
        self.area_mask_builder.fit(graph)
        return super().register_nodes(graph)


class TriNodes(IcosahedralNodes):
    """Nodes based on iterative refinements of an icosahedron.

    It depends on the trimesh Python library.
    """

    multi_scale_edge_cls: str = "anemoi.graphs.generate.multi_scale_edges.TriNodesEdgeBuilder"

    def create_nodes(self) -> tuple[nx.Graph, np.ndarray, list[int]]:
        from anemoi.graphs.generate.tri_icosahedron import create_tri_nodes

        return create_tri_nodes(resolution=max(self.resolutions))


class HexNodes(IcosahedralNodes):
    """Nodes based on iterative refinements of an icosahedron.

    It depends on the h3 Python library.
    """

    multi_scale_edge_cls: str = "anemoi.graphs.generate.multi_scale_edges.HexNodesEdgeBuilder"

    def create_nodes(self) -> tuple[nx.Graph, np.ndarray, list[int]]:
        from anemoi.graphs.generate.hex_icosahedron import create_hex_nodes

        return create_hex_nodes(resolution=max(self.resolutions))


class LimitedAreaTriNodes(LimitedAreaIcosahedralNodes):
    """Nodes based on iterative refinements of an icosahedron using an area of interest.

    It depends on the trimesh Python library.

    Parameters
    ----------
    area_mask_builder: KNNAreaMaskBuilder
        The area of interest mask builder.
    """

    multi_scale_edge_cls: str = "anemoi.graphs.generate.multi_scale_edges.TriNodesEdgeBuilder"

    def create_nodes(self) -> tuple[nx.Graph, np.ndarray, list[int]]:
        from anemoi.graphs.generate.tri_icosahedron import create_tri_nodes

        return create_tri_nodes(resolution=max(self.resolutions), area_mask_builder=self.area_mask_builder)


class LimitedAreaHexNodes(LimitedAreaIcosahedralNodes):
    """Nodes based on iterative refinements of an icosahedron using an area of interest.

    It depends on the h3 Python library.

    Parameters
    ----------
    area_mask_builder: KNNAreaMaskBuilder
        The area of interest mask builder.
    """

    multi_scale_edge_cls: str = "anemoi.graphs.generate.multi_scale_edges.HexNodesEdgeBuilder"

    def create_nodes(self) -> tuple[nx.Graph, np.ndarray, list[int]]:
        from anemoi.graphs.generate.hex_icosahedron import create_hex_nodes

        return create_hex_nodes(resolution=max(self.resolutions), area_mask_builder=self.area_mask_builder)


class StretchedIcosahedronNodes(LimitedAreaIcosahedralNodes, ABC):
    """Nodes based on iterative refinements of an icosahedron with 2
    different resolutions.

    Attributes
    ----------
    area_mask_builder : KNNAreaMaskBuilder
        The area of interest mask builder.
    """

    def __init__(
        self,
        global_resolution: int,
        lam_resolution: int,
        name: str,
        reference_node_name: str,
        mask_attr_name: str | None = None,
        margin_radius_km: float = 100.0,
    ) -> None:
        super().__init__(
            resolution=lam_resolution,
            reference_node_name=reference_node_name,
            mask_attr_name=mask_attr_name,
            margin_radius_km=margin_radius_km,
            name=name,
        )
        self.global_resolution = global_resolution


class StretchedTriNodes(StretchedIcosahedronNodes):
    """Nodes based on iterative refinements of an icosahedron with 2
    different resolutions.

    It depends on the trimesh Python library.
    """

    multi_scale_edge_cls: str = "anemoi.graphs.generate.multi_scale_edges.StretchedTriNodesEdgeBuilder"

    def create_nodes(self) -> tuple[nx.Graph, np.ndarray, list[int]]:
        from anemoi.graphs.generate.tri_icosahedron import create_stretched_tri_nodes

        return create_stretched_tri_nodes(
            base_resolution=self.global_resolution,
            lam_resolution=max(self.resolutions),
            area_mask_builder=self.area_mask_builder,
        )



# (C) Copyright 2024-2025 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import logging
from pathlib import Path  # noqa: TC003
from typing import Annotated
from typing import Literal
from typing import Optional

from pydantic import Field
from pydantic import PositiveFloat
from pydantic import PositiveInt

from anemoi.utils.schemas import BaseModel

LOGGER = logging.getLogger(__name__)


class AnemoiDatasetNodeSchema(BaseModel):
    target_: Literal["anemoi.graphs.nodes.AnemoiDatasetNodes", "anemoi.graphs.nodes.ZarrDatasetNodes"] = Field(
        ..., alias="_target_"
    )
    "Nodes from Anemoi dataset class implementation from anemoi.graphs.nodes."
    dataset: str | list | dict  # TODO(Helen): Discuss schema with Baudouin
    "The dataset containing the nodes."


class NPZnodeSchema(BaseModel):
    target_: Literal["anemoi.graphs.nodes.NPZFileNodes"] = Field(..., alias="_target_")
    "Nodes from NPZ grids class implementation from anemoi.graphs.nodes."
    npz_file: str
    "Path to the NPZ file."
    lat_key: str = Field(examples=["lat", "latitude"])
    "The key name of the latitude field."
    lon_key: str = Field(examples=["lon", "longitude"])
    "The key name of the longitude field."


class TextNodeSchema(BaseModel):
    target_: Literal["anemoi.graphs.nodes.TextNodes"] = Field(..., alias="_target_")
    "Nodes from text file class implementation from anemoi.graphs.nodes."
    dataset: str | Path
    "The path to text file containing the coordinates of the nodes."
    idx_lon: int
    "The index of the longitude in the dataset."
    idx_lat: int
    "The index of the latitude in the dataset."


class XArrayNodeSchema(BaseModel):
    target_: Literal["anemoi.graphs.nodes.XArrayNodes"] = Field(..., alias="_target_")
    "Nodes from xarray dataset class implementation from anemoi.graphs.nodes."
    dataset: str | Path
    "The path to xarray dataset containing the coordinates of the nodes."
    lon_key: str
    "The key name of the longitude field."
    lat_key: str
    "The key name of the latitude field."


class ReducedGaussianGridNodeSchema(BaseModel):
    target_: Literal["anemoi.graphs.nodes.ReducedGaussianGridNodes"] = Field(..., alias="_target_")
    "Nodes from NPZ grids class implementation from anemoi.graphs.nodes."
    grid: Literal["o16", "o32", "o48", "o96", "o160", "o256", "o320", "n320", "o1280"]
    "Reduced gaussian grid."


class ICONMeshNodeSchema(BaseModel):
    target_: Literal[
        "anemoi.graphs.nodes.ICONMultiMeshNodes",
        "anemoi.graphs.nodes.ICONCellGridNodes",
    ] = Field(
        ...,
        alias="_target_",
    )
    "Mesh based on ICON grid class implementation from anemoi.graphs.nodes."
    grid_filename: str
    "Name of NetCDF ICON grid file."
    max_level: int
    "Maximum refinement level of the multi mesh / cell grid."


class LimitedAreaNPZFileNodesSchema(BaseModel):
    target_: Literal["anemoi.graphs.nodes.LimitedAreaNPZFileNodes"] = Field(..., alias="_target_")
    "Class implementation for nodes from NPZ grids using an area of interest from anemoi.graphs.nodes."
    grid_definition_path: str
    "Path to the folder containing the grid definition files."
    resolution: str = Field(examples=["o48", "o96", "n320"])  # TODO(Helen): Discuss with Mario about the refactor
    "The grid resolution."
    reference_node_name: str  # TODO(Helen): Check that reference nodes exists in the config
    "Name of the reference nodes in the graph to consider for the Area Mask."
    mask_attr_name: Optional[str]  # TODO(Helen): Check that mask_attr_name exists in the dataset config
    "Name of a node to attribute to mask the reference nodes, if desired. Defaults to consider all reference nodes."
    margin_radius_km: PositiveFloat = Field(example=100.0)
    "Maximum distance to the reference nodes to consider a node as valid, in kilometers. Defaults to 100 km."


class IcosahedralandHealPixNodeSchema(BaseModel):
    target_: Literal[
        "anemoi.graphs.nodes.TriNodes",
        "anemoi.graphs.nodes.HexNodes",
        "anemoi.graphs.nodes.HEALPixNodes",
    ] = Field(..., alias="_target_")
    "Icohedral and HEAL Pix nodes class implementations from anemoi.graphs.nodes."
    resolution: PositiveInt
    "Refinement level of the mesh."


class LimitedAreaIcosahedralandHealPixNodeSchema(BaseModel):
    target_: Literal[
        "anemoi.graphs.nodes.LimitedAreaTriNodes",
        "anemoi.graphs.nodes.LimitedAreaHexNodes",
        "anemoi.graphs.nodes.LimitedAreaHEALPixNodes",
    ] = Field(..., alias="_target_")
    "Class implementations for Icosahedral and HEAL Pix nodes using an area of interest from anemoi.graphs.nodes."
    resolution: PositiveInt
    "Refinement level of the mesh."
    reference_node_name: str  # TODO(Helen): Discuss check that reference nodes exists in the config
    "Name of the reference nodes in the graph to consider for the Area Mask."
    mask_attr_name: Optional[str]  # TODO(Helen): Discuss check that mask_attr_name exists in the dataset config
    "Name of a node to attribute to mask the reference nodes, if desired. Defaults to consider all reference nodes."
    margin_radius_km: PositiveFloat = Field(example=100.0)
    "Maximum distance to the reference nodes to consider a node as valid, in kilometers. Defaults to 100 km."


class StretchedIcosahdralNodeSchema(BaseModel):
    target_: Literal["anemoi.graphs.nodes.StretchedTriNodes"] = Field(..., alias="_target_")
    "Class implementation for nodes based on iterative refinements of an icosahedron with 2 different resolutions."
    global_resolution: PositiveInt
    "Refinement level of the mesh on the global area."
    lam_resolution: PositiveInt
    "Refinement level of the mesh on the local area."
    reference_node_name: str
    "Name of the reference nodes in the graph to consider for the Area Mask."
    mask_attr_name: Optional[str]
    "Name of a node to attribute to mask the reference nodes, if desired. Defaults to consider all reference nodes."
    margin_radius_km: PositiveFloat = Field(example=100.0)
    "Maximum distance to the reference nodes to consider a node as valid, in kilometers. Defaults to 100 km."


NodeBuilderSchemas = Annotated[
    AnemoiDatasetNodeSchema
    | NPZnodeSchema
    | TextNodeSchema
    | ICONMeshNodeSchema
    | LimitedAreaNPZFileNodesSchema
    | ReducedGaussianGridNodeSchema
    | IcosahedralandHealPixNodeSchema
    | LimitedAreaIcosahedralandHealPixNodeSchema
    | StretchedIcosahdralNodeSchema,
    Field(discriminator="target_"),
]
